/*
 * Hexagonal Grid
 * Author: niasw
 */

#include "hexgrid.hpp"
using namespace hexagon;

bool veci::initOK = 0;
veci* veci::adj = 0;
veci veci::adjstart(1,0,-1);

/*---vector integer---*/
veci::veci(int x1,int x2,int x3)
{
    x=new int[3];
    x[0]=x1;
    x[1]=x2;
    x[2]=x3;
    if (!initOK)
    {
        initOK = true;
        //start = (1,0,-1)
        adj = new veci[6];
        adj[0]=veci(-1,1,0);
        adj[1]=veci(-1,0,1);
        adj[2]=veci(0,-1,1);
        adj[3]=veci(1,-1,0);
        adj[4]=veci(1,0,-1);
        adj[5]=veci(0,1,-1);
    }
}
veci::veci(const veci& vy)
{
    x=new int[3];
    x[0]=vy[0];
    x[1]=vy[1];
    x[2]=vy[2];
}
veci::~veci()
{
    delete[] x;
}

veci& veci::operator= (const veci& vy)
{
    x[0]=vy[0];
    x[1]=vy[1];
    x[2]=vy[2];
    return *this;
}
veci& veci::operator+= (const veci& vy)
{
    x[0]+=vy[0];
    x[1]+=vy[1];
    x[2]+=vy[2];
    return *this;
}
veci& veci::operator-= (const veci& vy)
{
    x[0]-=vy[0];
    x[1]-=vy[1];
    x[2]-=vy[2];
    return *this;
}

double veci::d() const
{
    return sqrt(d2());
}

double hexagon::d2(const veci& v)
{
    return (v[0]*v[0]+v[1]*v[1]+v[2]*v[2])/2.;
}
double hexagon::d(const veci& v)
{
    return sqrt(d2(v));
}

std::ostream& hexagon::operator<< (std::ostream& os, const veci& vo)
{
    return os<<'<'<<vo[0]<<','<<vo[1]<<','<<vo[2]<<'>';
}
veci hexagon::operator* (const int k, const veci& v)
{
    return veci(v[0]*k,v[1]*k,v[2]*k);
}

/*---vector double---*/
vecd::vecd(double x1,double x2,double x3)
{
    x=new double[3];
    x[0]=x1;
    x[1]=x2;
    x[2]=x3;
}
vecd::vecd(double xd,double yd)
{
    x=new double[3];
    x[0]=xd;
    x[1]=(sqrt(3)*yd-xd)/2;
    x[2]=-(sqrt(3)*yd+xd)/2;
}
vecd::vecd(const vecd& vy)
{
    x=new double[3];
    x[0]=vy[0];
    x[1]=vy[1];
    x[2]=vy[2];
}
vecd::vecd(const veci& vy)
{
    x=new double[3];
    x[0]=vy[0];
    x[1]=vy[1];
    x[2]=vy[2];
}
vecd::~vecd()
{
    delete[] x;
}

vecd& vecd::operator= (const vecd& vy)
{
    x[0]=vy[0];
    x[1]=vy[1];
    x[2]=vy[2];
    return *this;
}
vecd& vecd::operator+= (const vecd& vy)
{
    x[0]+=vy[0];
    x[1]+=vy[1];
    x[2]+=vy[2];
    return *this;
}
vecd& vecd::operator-= (const vecd& vy)
{
    x[0]-=vy[0];
    x[1]-=vy[1];
    x[2]-=vy[2];
    return *this;
}

double vecd::d() const
{
    return sqrt(d2());
}

double hexagon::d2(const vecd& v)
{
    return (v[0]*v[0]+v[1]*v[1]+v[2]*v[2])/2.;
}
double hexagon::d(const vecd& v)
{
    return sqrt(d2(v));
}

std::ostream& hexagon::operator<< (std::ostream& os, const vecd& vo)
{
    return os<<'<'<<vo[0]<<','<<vo[1]<<','<<vo[2]<<'>';
}
vecd hexagon::operator* (const double k, const vecd& v)
{
    return vecd(v[0]*k,v[1]*k,v[2]*k);
}

void hexagon::hexagon2square(double* xy, const vecd& v)
{
    xy[0]=v[0];
    xy[1]=(v[1]-v[2])/sqrt(3);
}

hexagondev::hexagondev(int sidelength)
{
    if (sidelenth<0) sidelenth=0;
    sidelenth=sidelength;
    nodes_num=sidelength;
    ++sidelength;
    nodes_num*=sidelength;
    nodes_num/=2;
    nodes=new veci[6*nodes_num+1];
    int t; //triangle partion
    int p1,p2; //index position
    int s1,s2; //index signal
    int order=1;
    int t1,t2; //iterators
    for (t=0;t<6;++t)
    {
        switch (t)
        {
            case 0: //partion 1
                p1=0;p2=2;
                s1=+1;s2=-1;
                break;
            case 1: //partion 2
                p1=2;p2=1;
                s1=-1;s2=+1;
                break;
            case 2: //partion 3
                p1=1;p2=0;
                s1=+1;s2=-1;
                break;
            case 3: //partion 4
                p1=0;p2=2;
                s1=-1;s2=+1;
                break;
            case 4: //partion 5
                p1=2;p2=1;
                s1=+1;s2=-1;
                break;
            case 5: //partion 6
                p1=1;p2=0;
                s1=-1;s2=+1;
                break;
        };
        for (t1=1;t1<=sidelenth;++t1)
        {
            for (t2=0;t2<t1;++t2)
            {
                nodes[order][p1]=t1*s1;
                nodes[order][p2]=t2*s2;
                nodes[order][3-p1-p2]=-t1*s1-t2*s2;
                ++order;
            }
        }
    }
}

hexagondev::~hexagondev()
{
    delete[] nodes;
}

void hexagondev::resize(int sidelength)
{
    if (sidelenth<0)
    {
        std::cerr<<"Invalid Size."<<std::endl;
        return;
    }
    delete[] nodes;
    sidelenth=sidelength;
    nodes_num=sidelength;
    ++sidelength;
    nodes_num*=sidelength;
    nodes_num/=2;
    nodes=new veci[6*nodes_num+1];
    int t; //triangle partion
    int p1,p2; //index position
    int s1,s2; //index signal
    int order=1;
    int t1,t2; //iterators
    for (t=0;t<6;++t)
    {
        switch (t)
        {
            case 0: //partion 1
                p1=0;p2=2;
                s1=+1;s2=-1;
                break;
            case 1: //partion 2
                p1=2;p2=1;
                s1=-1;s2=+1;
                break;
            case 2: //partion 3
                p1=1;p2=0;
                s1=+1;s2=-1;
                break;
            case 3: //partion 4
                p1=0;p2=2;
                s1=-1;s2=+1;
                break;
            case 4: //partion 5
                p1=2;p2=1;
                s1=+1;s2=-1;
                break;
            case 5: //partion 6
                p1=1;p2=0;
                s1=-1;s2=+1;
                break;
        };
        for (t1=1;t1<=sidelenth;++t1)
        {
            for (t2=0;t2<t1;++t2)
            {
                nodes[order][p1]=t1*s1;
                nodes[order][p2]=t2*s2;
                nodes[order][3-p1-p2]=-t1*s1-t2*s2;
                ++order;
            }
        }
    }
}

int hexagondev::operator[] (const veci& vector) const
{
    if (!vector.check())
    {
        return -1; //invalid vector
    }
    int t,p1,p2,s1,t1,t2; //see the construction function of 'hexagondev'
    t1=0;
    p1=-1;///debug
    for (int it=0;it<3;++it)
    {
        if (abs(vector[it])>t1)
        {
            p1=it;
            s1=(vector[it]>0); //1 for +, 0 for -
            t1=abs(vector[it]);
        }
    }
    if (t1==0)
    {
        return 0; //<0,0,0>
    }
    if (t1>sidelenth)
    {
        return -1; //outside
    }
    if (p1==0 && abs(vector[2])==t1)
    {
        p1=2;
        s1=(vector[2]>0);
    }
    p2=(p1+2)%3;
    t2=abs(vector[p2]);
    t=2*p1;
    if (!s1)
    {
        ++t;
        t%=3;
        t=5-t*2;
    }
    return t*nodes_num+t1*(t1-1)/2+t2+1;
}
