#ifndef _OPENGL_
#define _OPENGL_
#include <GL/glut.h>
#endif

#include "hexgridplus.hpp"
using namespace hexagonplus;
#include <fstream>
using namespace std;

/*
 * Local API Functions
 */
static void reshape ( int width, int height );
static void display ();
static void idle ();
static void keyboard ( unsigned char key, int x, int y );
static void keyup ( unsigned char key, int x, int y );
static void timer ( int value );

static void GraphicSettings ();
static void DoBeforeExiting ();
static void KeyPressedEvent ();

int topwinhandle; //handle of Top Window
int subwinhandle[5]; //handle of Sub Window
int curwin=0; //current window
/*
 * Local Graphic Functions
 */
void glusrHexagon(double size);

/*
 * Local Project Functions
 */
void Initialization(); //load .rec file head

/*
 * Global Graphic Variables
 */
unsigned keypressed[256];

#define _LIGHT_
GLfloat mat_ambient_default[]    = { 1.0, 1.0, 1.0, 1.0 };
GLfloat mat_diffuse_default[]    = { 0.5, 0.5, 0.5, 1.0 };
GLfloat mat_specular_default[]   = { 0.7, 0.7, 0.7, 1.0 };
GLfloat mat_emission_default[]   = { 0.0, 0.0, 0.0, 1.0 };
GLfloat high_shininess_default[] = { 100.0 };

GLfloat mat_emission_cell[] = { 1.0, 0.0, 0.0, 1.0,
                                0.0, 1.0, 0.0, 1.0,
                                0.0, 0.0, 1.0, 1.0};

float slope_angle=45;
float precession =-90;
float viewer_size=0.1;

/*
 * Global Project Variables
 */

hexagondev dev(1); //contains the cells' world
int devSideLen; //side length of the device
double* var; //variables of the cell
double* MAXvar; //the upper limit to show var
double STAtime; //start time
double ENDtime; //end time
double time_step; //time step used in calculation
double nowtime; //time in calculation
int varnum; //number of variables

ifstream ifs; //loading stream
int startsite; //start site of the record data
const char* filename; //filename of the .rec file

int time_fresh=30; //fresh screen and evolve system
int play_speed=1; //number of frames per freshtime
int cur_frame=0; //current frame ID
int end_frame; //end frame ID (no data for the end frame)
int framespace; //Bytes per frame

/*
 * Temporary Variables
 */
 //temporary Descartes Coordinates
double ty1[2];
double ty2[2];
 //iterators
int it0,it1,it2,it3;
int tmp0,tmp1,tmp2;

int selected=0; //highlight the chosen
int timedisplay=0; //times of displaying
int surface=0; //main display
int play_direc=1; //playing direction

double* curvar; //point at current value

int winwidth, winheight; //current window size
double winar; //width/height

//--------------Main Procedure------------
int main ( int argc, char *argv[] )
{
    if (argc=1)
    {
        filename=argv[1];
    }
    else if (argc>1)
    {
        cout<<"Usage: DynamicView [inputfile.rec]"<<endl;
        return 0;
    }
    else
    {
        filename="SimuData.rec";
    }

    glutInit ( &argc, argv );
    glutInitDisplayMode ( GLUT_RGBA | GLUT_DOUBLE | GLUT_DEPTH );
    glutInitWindowPosition ( 0, 0 );
    glutInitWindowSize ( 1366, 768 );
    //glutFullScreen();
    topwinhandle = glutCreateWindow ( argv[0] );
    subwinhandle[1] = glutCreateSubWindow ( topwinhandle, 0, 0, 384, 384 );
    subwinhandle[2] = glutCreateSubWindow ( topwinhandle, 0, 384, 384, 384 );
    subwinhandle[3] = glutCreateSubWindow ( topwinhandle, 982, 0, 384, 384 );
    subwinhandle[4] = glutCreateSubWindow ( topwinhandle, 982, 384, 384, 384 );
    subwinhandle[0] = glutCreateSubWindow ( topwinhandle, 384, 0, 598, 768 );

    glutReshapeFunc ( reshape );
    glutDisplayFunc ( display );
    glutIdleFunc ( idle );
    glutTimerFunc ( time_fresh, timer, 0 );
    glutKeyboardFunc ( keyboard );
    glutKeyboardUpFunc ( keyup );

    GraphicSettings ();

    Initialization ();
    glutMainLoop ();
    return 0;
}

void GraphicSettings ()
{
  for (it0=0;it0<5;++it0)
  {
    glutSetWindow ( subwinhandle[it0] );

    glClearColor ( 0, 0, 0, 1 );

    glutIgnoreKeyRepeat (1);

    glPolygonMode ( GL_FRONT_AND_BACK, GL_FILL );

    glCullFace( GL_BACK );
    glEnable ( GL_CULL_FACE );
    glShadeModel (GL_SMOOTH);

    glEnable ( GL_DEPTH_TEST );
    glDepthFunc ( GL_LEQUAL );

    glEnable ( GL_NORMALIZE );
    glEnable ( GL_COLOR_MATERIAL );
    glEnable ( GL_LIGHTING );

    glEnable ( GL_BLEND );
    glBlendFunc ( GL_SRC_ALPHA , GL_ONE_MINUS_SRC_ALPHA  );

    glEnable ( GL_LINE_SMOOTH );
    glHint ( GL_LINE_SMOOTH_HINT , GL_FASTEST );

    glEnable ( GL_POLYGON_SMOOTH );
    glHint ( GL_POLYGON_SMOOTH_HINT , GL_FASTEST );

    glMaterialfv ( GL_FRONT_AND_BACK, GL_AMBIENT,   mat_ambient_default    );
    glMaterialfv ( GL_FRONT_AND_BACK, GL_DIFFUSE,   mat_diffuse_default    );
    glMaterialfv ( GL_FRONT_AND_BACK, GL_SPECULAR,  mat_specular_default   );
    glMaterialfv ( GL_FRONT_AND_BACK, GL_SHININESS, high_shininess_default );
  }

  for (int it=0; it<256; ++it) keypressed[it] = 0;
}

static void DoBeforeExiting ()
{
    delete[] var;
    delete[] MAXvar;

    ifs.close();
}

static void KeyPressedEvent ()
{
    if ( keypressed['w'] )
    {
       if (slope_angle>=1) slope_angle-=1;
    }
    if ( keypressed['s'] )
    {
       if (slope_angle<=89) slope_angle+=1;
    }
    if ( keypressed['d'] )
    {
       precession +=3;
       if (precession > 360) precession-=360;
    }
    if ( keypressed['a'] )
    {
       precession -=3;
       if (precession < 0) precession+=360;
    }
    if ( keypressed['e'] )
    {
       if (viewer_size<1) viewer_size*=1.2;
    }
    if ( keypressed['q'] )
    {
       if (viewer_size>0.01) viewer_size/=1.2;
    }
}

static void reshape ( int width, int height )
{
}

static void display ()
{
    glutSetWindow ( subwinhandle[curwin] );

    winwidth = glutGet ( GLUT_WINDOW_WIDTH );
    winheight = glutGet ( GLUT_WINDOW_HEIGHT );
    winar = double ( winwidth ) / double ( winheight );

    glViewport ( 0, 0, winwidth, winheight );

    glMatrixMode ( GL_PROJECTION );
    glLoadIdentity ();
    glOrtho ( -winar, winar, -1, 1, 0.5, 200 );

    glMatrixMode ( GL_MODELVIEW );
    glLoadIdentity ();

    glPushMatrix();
      glTranslatef(0,0,-10);
      glRotatef(-slope_angle,1,0,0);
      glRotatef(-precession,0,0,1);

    it2 = (curwin + surface) % varnum;
    curvar = var + it2 * dev.GetNodeNumber();

    glClear ( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );

    for (it1=0;it1<dev.GetNodeNumber();++it1)
    {
        mat_emission_cell[it2%3*5]=2./3*(0.5+curvar[it1]/MAXvar[it2]);
        if (it1!=selected) glMaterialfv ( GL_FRONT_AND_BACK, GL_EMISSION,  mat_emission_cell + it2 % 3 * 4 );
        else glMaterialfv ( GL_FRONT_AND_BACK, GL_EMISSION,  mat_ambient_default );
        hexagon2square(ty1,dev[it1]);
        glPushMatrix();
          glTranslatef(viewer_size*ty1[0],viewer_size*ty1[1],3*viewer_size*curvar[it1]/MAXvar[it2]);
          glusrHexagon(viewer_size);
        glPopMatrix();
    }

    glPopMatrix();

    glutSwapBuffers ();

    ++curwin;
    curwin%=5;
}

static void keyboard ( unsigned char key, int x, int y )
{
    if ( key == 27 ) //Exit
    {
        DoBeforeExiting();
        exit(0);
    }
    if ( key == 'w' )
    {
        keypressed['w'] = 1;
    }
    if ( key == 's' )
    {
        keypressed['s'] = 1;
    }
    if ( key == 'd' )
    {
        keypressed['d'] = 1;
    }
    if ( key == 'a' )
    {
        keypressed['a'] = 1;
    }
    if ( key == 'e' )
    {
        keypressed['e'] = 1;
    }
    if ( key == 'q' )
    {
        keypressed['q'] = 1;
    }
    if ( key == 'u' )
    {
        ++surface;
        surface%=varnum;
    }
    if ( key == 't' )
    {
        int tmp=dev[dev[selected]+veci(1,0,-1)];
        if (tmp>=0) selected=tmp;
    }
    if ( key == 'f' )
    {
        int tmp=dev[dev[selected]+veci(0,1,-1)];
        if (tmp>=0) selected=tmp;
    }
    if ( key == 'v' )
    {
        int tmp=dev[dev[selected]+veci(-1,1,0)];
        if (tmp>=0) selected=tmp;
    }
    if ( key == 'b' )
    {
        int tmp=dev[dev[selected]+veci(-1,0,1)];
        if (tmp>=0) selected=tmp;
    }
    if ( key == 'h' )
    {
        int tmp=dev[dev[selected]+veci(0,-1,1)];
        if (tmp>=0) selected=tmp;
    }
    if ( key == 'y' )
    {
        int tmp=dev[dev[selected]+veci(1,-1,0)];
        if (tmp>=0) selected=tmp;
    }
    if ( key == 'g' )
    {
        std::cout<<std::endl;
        std::cout<<"order="<<selected<<std::endl;
        std::cout<<"coord="<<dev[selected]<<std::endl;
        std::cout<<"curvar="<<var[selected+surface*dev.GetNodeNumber()]<<std::endl;
    }
    if ( key == 0x20 )
    {
        if (play_speed!=0) play_speed=0;
        else
        {
            play_speed=1;
            play_direc=1;
        }
    }
    if ( key == 'r' )
    {
        cur_frame=0;
        play_speed=0;
    }
    if ( key == 'i' )
    {
        play_direc=-1;
    }
    if ( key == ']' )
    {
        play_speed*=2;
    }
    if ( key == '[' )
    {
        if (play_speed>1) play_speed/=2;
    }
}

static void keyup ( unsigned char key, int x, int y )
{
    if ( key == 'w' )
    {
        keypressed['w'] = 0;
    }
    if ( key == 's' )
    {
        keypressed['s'] = 0;
    }
    if ( key == 'd' )
    {
        keypressed['d'] = 0;
    }
    if ( key == 'a' )
    {
        keypressed['a'] = 0;
    }
    if ( key == 'e' )
    {
        keypressed['e'] = 0;
    }
    if ( key == 'q' )
    {
        keypressed['q'] = 0;
    }
}

static void idle ()
{
    glutPostRedisplay ();
}

static void timer ( int value )
{
  if (curwin==0)
  {
    ifs.read((char*)var,framespace);
    cur_frame+=play_speed*play_direc;
    if (cur_frame >= end_frame) cur_frame-=play_speed;
    if (cur_frame < 0) cur_frame+=play_speed;
    ifs.seekg(startsite + cur_frame * framespace, ios::beg );

    for (it1=0;it1<varnum;++it1)
    {
        MAXvar[it1]=0;
        curvar = var + it1 * dev.GetNodeNumber();
        for (it2=0;it2<dev.GetNodeNumber();++it2)
        {
            if (curvar[it2]>MAXvar[it1])
            {
                MAXvar[it1]=curvar[it2];
            }
        }
    }

    cout<<"\xd"<<"Current Frame: "<<cur_frame<<"    "<<"\tTime: "<<STAtime + cur_frame * time_step<<"    "<<flush;
  }
  KeyPressedEvent ();
  glutPostRedisplay ();
  glutTimerFunc ( time_fresh, timer, 0 );
}

void glusrHexagon(double size)
{
    double ty[2];
    glPushMatrix();
      glRotatef(30,0,0,1);
      glBegin(GL_POLYGON);
      for (it3=0;it3<6;++it3)
      {
          hexagon2square(ty,veci::adj[it3]);
          glVertex3d(ty[0]*0.6*size,ty[1]*0.6*size,0);
      }
      glEnd();
    glPopMatrix();
}

void Initialization()
{
    ifs.open(filename,ios::binary);

    ifs.read((char*)&it0,4);
    if (it0 & 0x00ffffff != 0x584548)
    {
        cerr<<"Invalid .rec file."<<endl;
        exit(1);
    }
    ifs.read((char*)&startsite,4);
    ifs.read((char*)&devSideLen,4);
    cout<<"Device Side Length: "<<devSideLen<<endl;
    ifs.read((char*)&varnum,4);
    cout<<"Number of Variables: "<<varnum<<endl;
    ifs.read((char*)&STAtime,8);
    cout<<"Time Start: "<<STAtime<<endl;
    ifs.read((char*)&ENDtime,8);
    cout<<"Time Final: "<<ENDtime<<endl;
    ifs.read((char*)&time_step,8);
    cout<<"Time/Frame: "<<time_step<<endl;
    ifs.read((char*)&it0,4);
    cout<<"Calc/Save : "<<it0<<endl; //calculating frames per viewing frame

    dev.resize(devSideLen);
    end_frame = (ENDtime-STAtime)/time_step;
    framespace = dev.GetNodeNumber() * varnum * 8;
    var=new double[varnum*dev.GetNodeNumber()];
    MAXvar=new double[varnum];
    cout<<"Total Frames: "<<end_frame<<endl;
    cout<<framespace<<" Bytes / Frame"<<endl;
    ifs.seekg(startsite, ios::beg);
    cout<<"Current Frame: "<<0<<flush;
}

